# pylint: disable=R0801
# pylint: disable=missing-function-docstring
"""
Test Input Status.
"""
import asyncio
import json
import logging
import unittest

from unii import UNiiInputState, UNiiLocal

_SETTINGS_JSON = "settings.json"

_LOGGER = logging.getLogger(__name__)


class Test(unittest.IsolatedAsyncioTestCase):
    """
    Test Input Status.
    """

    _unii = None

    async def asyncSetUp(self):
        with open(_SETTINGS_JSON, encoding="utf8") as settings_file:
            settings = json.load(settings_file)
            host = settings.get("host")
            port = settings.get("encrypted_port", 6502)
            shared_key = settings.get("shared_key", 6502)
            self.user_code = settings.get("user_code", "123456")

            self._unii = UNiiLocal(host, port, shared_key)
            await asyncio.sleep(1)
            result = await self._unii.connect()
            self.assertTrue(result, "Failed to connect to UNii")

    async def asyncTearDown(self):
        await asyncio.sleep(1)
        await self._unii.disconnect()

    async def test_inputs(self):
        """
        Test connecting to Alphatronics UNii.
        """
        self.assertIsNotNone(self._unii.inputs, "Input Status not set")
        for _, unii_input in self._unii.inputs.items():
            if unii_input.status != UNiiInputState.DISABLED:
                _LOGGER.info(unii_input)

    async def test_bypass_wired_input(self):
        await self._unii.unbypass_input(2, self.user_code)
        result = await self._unii.bypass_input(2, self.user_code)
        self.assertTrue(result, "Failed to bypass input")

    async def test_bypass_wired_input_fail(self):
        await self._unii.unbypass_input(2, self.user_code)
        result = await self._unii.bypass_input(2, "")
        self.assertFalse(result)

    async def test_unbypass_wired_input(self):
        await self._unii.bypass_input(2, self.user_code)
        result = await self._unii.unbypass_input(2, self.user_code)
        self.assertTrue(result, "Failed to bypass input")

    async def test_unbypass_wired_input_fail(self):
        await self._unii.bypass_input(2, self.user_code)
        result = await self._unii.unbypass_input(2, "")
        self.assertFalse(result)

    async def test_bypass_keypad_input(self):
        await self._unii.unbypass_input(701, self.user_code)
        result = await self._unii.bypass_input(701, self.user_code)
        self.assertTrue(result, "Failed to bypass input")

    async def test_bypass_keypad_input_fail(self):
        await self._unii.unbypass_input(701, self.user_code)
        result = await self._unii.bypass_input(701, "")
        self.assertFalse(result)



# pylint: disable=R0801
# pylint: disable=missing-function-docstring
"""
Test creating UNii messages.
"""

import logging
import unittest

from unii.sia_code import SIACode
from unii.unii_command import UNiiCommand
from unii.unii_message import UNiiChecksumError, UNiiRequestMessage, UNiiResponseMessage

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Test UNii message.
    """

    def test_unencrypted_connection_request(self):
        """
        Creates an unencrypted connection request message and validates it's correctness.
        """
        message = UNiiRequestMessage()
        message.session_id = 0xFFFF
        message.tx_sequence = 0x08BE2C53
        message.rx_sequence = 0x00000000
        message.command = UNiiCommand.CONNECTION_REQUEST
        self.assertEqual(
            message.to_bytes().hex(),
            "ffff08be2c530000000004010020000100000000000000000000000000005ed3",
        )

    def test_unencrypted_connection_response(self):
        """
        Creates an unencrypted connection response message and validates it's correctness.
        """
        message = UNiiResponseMessage(
            bytes.fromhex("f109441a389608be2c530402001400020000853e")
        )
        self.assertEqual(message.session_id, 0xF109)
        self.assertEqual(message.tx_sequence, 0x441A3896)
        self.assertEqual(message.rx_sequence, 0x08BE2C53)
        self.assertEqual(message.command, UNiiCommand.CONNECTION_REQUEST_RESPONSE)
        self.assertIsNone(message.data)

    def test_unencrypted_connection_response_wrong_checksum(self):
        """
        Creates an unencrypted connection response message with an invallid checksum.

        Should throw an UNiiChecksumError
        """
        self.assertRaises(
            UNiiChecksumError,
            UNiiResponseMessage,
            bytes.fromhex("f109441a389608be2c530402001400020000853d"),
        )

    def test_unencrypted_event_occurred(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "b11177c121e200007ca20402004501020031000381b17c030b021d1116436f6e66696775726174696520676577696a7a696764000000000000000000000000000020204a12"
            )
        )
        self.assertEqual(message.session_id, 0xB111)
        self.assertEqual(message.tx_sequence, 0x77C121E2)
        self.assertEqual(message.rx_sequence, 0x00007CA2)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Configuratie gewijzigd")
        self.assertIsNone(message.data.sia_code)

    def test_encrypted_event_occurred_1(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "22f816f54ec00000d64405020043bff2b8de0eda43d9ce70ca2ee5db0e5822f33b013682af245fb4d693d4620906bb681a26cde01779ad351de49b7ee0afa0e91492e6"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0x22F8)
        self.assertEqual(message.tx_sequence, 0x16F54EC0)
        self.assertEqual(message.rx_sequence, 0x0000D644)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Brand Alarm")
        self.assertEqual(message.data.sia_code, SIACode.FIRE_ALARM)

    def test_encrypted_event_occurred_2(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "22f816f54ec00000d64405020043bff2b8de0eda43d9ce70ca2ee5db0e5822f33b013682af245fb4d693d4620906bb681a26cde01779ad351de49b7ee0afa0e91492e6"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0x22F8)
        self.assertEqual(message.tx_sequence, 0x16F54EC0)
        self.assertEqual(message.rx_sequence, 0x0000D644)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Brand Alarm")
        self.assertEqual(message.data.sia_code, SIACode.FIRE_ALARM)

    def test_encrypted_event_occurred_3(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "ed6c37c13cfb0000a9c905020055ed4b1fa1b14709845e8895825e64005d3cc502c5a54c7f16ae04e7525ca1c3e895ce2b8d5e1125d818cade5d5f778b0cd22c7476d510c3194644b76d79e6fd8d9be76a73302236"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0xED6C)
        self.assertEqual(message.tx_sequence, 0x37C13CFB)
        self.assertEqual(message.rx_sequence, 0x0000A9C9)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Sabotage schakelaar alarm")
        self.assertEqual(message.data.device_number, 5016)
        self.assertEqual(message.data.device_name, "UNii keypad 1")
        self.assertEqual(message.data.sia_code, SIACode.TAMPER)

    def test_encrypted_event_occurred_4(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "bdc57b6ed1f9000038740502005e023995b766cc5cbcaf338306bc9adeedba83f1e4f04d720941ad52ef4b45fda57abd2115bd09018bd78b637a9ac55d7a985b69ffecbc334dac4b0e317c0503a7abe2cd85b789c8d85b0806e33f169c7a"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0xBDC5)
        self.assertEqual(message.tx_sequence, 0x7B6ED1F9)
        self.assertEqual(message.rx_sequence, 0x00003874)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Sabotage schakelaar herstel")
        self.assertEqual(message.data.device_number, 5016)
        self.assertEqual(message.data.device_name, "UNii keypad 1")
        self.assertEqual(message.data.sia_code, SIACode.TAMPER_RESTORAL)

    def test_encrypted_event_occurred_5(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "ed6c37c13cfb0000a9ca0502004d9350e2cae2ee6eb18dd9ae3aca0f3f073189fb6c1c0df68b7aada71d33985b3cd788853a02e9248ab2b0cb25fc1c6a59e40a1f99e2390ef6ac2f3a5e792e5b"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0xED6C)
        self.assertEqual(message.tx_sequence, 0x37C13CFB)
        self.assertEqual(message.rx_sequence, 0x0000A9CA)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Communicatie fout")
        self.assertEqual(message.data.device_number, 5016)
        self.assertEqual(message.data.device_name, "UNii keypad 1")
        self.assertEqual(message.data.sia_code, SIACode.TAMPER)

    def test_encrypted_event_occurred_6(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "bdc57b6ed1f900003873050200575df1479ce5177f839875bacdee013675d70e1a9ddc9b116e81d3f0986fdfbd8feb3b98fd198df3fb9b013d2a69295dbba95d2eaaa2c80dc07e776e6be73abd28c0ac765be39b498ffd"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0xBDC5)
        self.assertEqual(message.tx_sequence, 0x7B6ED1F9)
        self.assertEqual(message.rx_sequence, 0x00003873)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Communicatie herstel")
        self.assertEqual(message.data.device_number, 5016)
        self.assertEqual(message.data.device_name, "UNii keypad 1")
        self.assertEqual(message.data.sia_code, SIACode.TAMPER_RESTORAL)

    def test_encrypted_event_occurred_7(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "bdc57b6ed1f90000386b05020057c70b74b84c85b840568667e0054f03feeda2eff0eff4f1a57f98b9c3957896c855a6ab068755f675dcf0079c4b68633819cc67223099a1b54e652e7d5a7d1fa93a5628520d2dc54e0d"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0xBDC5)
        self.assertEqual(message.tx_sequence, 0x7B6ED1F9)
        self.assertEqual(message.rx_sequence, 0x0000386B)
        self.assertEqual(message.command, UNiiCommand.EVENT_OCCURRED)
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data.event_description, "Config gewijzigd op afstand")
        self.assertEqual(message.data.device_number, 5000)
        self.assertEqual(message.data.device_name, "UNii centrale")
        self.assertEqual(message.data.sia_code, SIACode.REMOTE_PROGRAM_SUCCESS)

    def test_section_arrangement(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "a6947bb0970e0000c78e05020040308e8d444538679e74bfde534279bf107e1df375b473bb91c40cd5b99d9316491e5b4b43cb806c8520cdb12660359db150c2"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0xA694)
        self.assertEqual(message.tx_sequence, 0x7BB0970E)
        self.assertEqual(message.rx_sequence, 0x0000C78E)
        self.assertEqual(
            message.command, UNiiCommand.RESPONSE_REQUEST_SECTION_ARRANGEMENT
        )
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data[1].active, True)
        self.assertEqual(message.data[1].name, "Sectie 1")

    def test_response_request_input_arrangement(self):
        """
        Test special characters in input name.
        """
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "3fe151586c9f00007b21050202085a28f03ab0176ace4a6ad46d78ffdfbe3fbb787deb4900641c76fdac26b8dc05f8caadd668b29039a76e284e87d949ce0fa53e899b06ad0aaf66605129521d82c141c67091fb0f48c3cadebc960313ece81fef4772c810d42f7f6d092ffcf04f842ce01f7acb1f3fa529eafc1ec2e1f6502667ea35396df1bfdda867fabe17f6c46cc120e251112c50cbefb33ecd1aa64923116b4eedda480f2120ea8c65d98b68131694c1dffd0060b9ab7845036352a184ec617aa35036e3ef4d0fef1db6710e8322b00b1a90fc576b4c904ac1b9e4e5821df26ab6e6b5c879047a24eeaf12a77acedceaa7d27531572d4fa7d81363795610df521ec62082f78fb77f453978e08d4ee85a9d1528fd79960a600e869f2c041474fdb77ec05dbd207ad73782df904c6c52a03101dbf22117d283936f07dc806090c5e7a1b9e64850a663a5a0585a473ccda610bd09b3dcc5727de5c708cc8166a0b39c89dd578ed88a3393c516bf9ff12c7c8fedab05ce47584dee193e15cb7bac2b31ed7e9c2b9e24cb11c3613b82a191ebf8666c15fd9e947612cc7e09e1abedeaa7fa8b7e56412798c97c3266b2cdebf9808fe65fe58abeca61479cf6bc62ac76e54dee5a9773d128798960352b5cc5fd7dbed0502a852f7576dc6172dcb54a2cb40a65b28c79fbb5c81a255c32f5305e3934b9925a99a584b087bd0c3d60e51eee79541695"
            ),
            bytes.fromhex("30313233343536373839616263646566"),
        )
        _LOGGER.debug(message)
        self.assertEqual(message.session_id, 0x3FE1)
        self.assertEqual(message.tx_sequence, 0x51586C9F)
        self.assertEqual(message.rx_sequence, 0x00007B21)
        self.assertEqual(
            message.command, UNiiCommand.RESPONSE_REQUEST_INPUT_ARRANGEMENT
        )
        self.assertIsNotNone(message.data)
        self.assertEqual(message.data[1].name, "Ingang 1 é ë ê 1ï  B")

    def test_response_request_output_arrangement(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex(
                "03a7155e79ad000093570502017d63849572e1395027d3310f4f6bdecb825097a6c2a28a03aad4a22985b0672d96513e390ab0c510e4875d00c875079f2a9d45cd6183f6ec632a57639336f7bfcb84074321c3cdff270347516063d03c6d4e8ae2e548e31d5726cc44d3361dea3075ec2ce2b826a8083778e1b95d244e7861b6f0abcbc24cdef47dc1f65a0f426fbf287090280eaf91ad21c54ebac2442b2827c805a37b3e4acd90431e4ab159be08ae713b873fe7c0c83b5c795f08543ee7d9439e3b7bebe9e66f0b3fea9aa9e0ac023e1da64c1a95a71ac93e82a4a78dc7b29f5325a3ccfc2f5420167ac4c4c0c344da7d05a202f921fb9491e155a822f891acb046eabb323459393e1e2b1746ca4bf1d9f93097d1560dd50eed4870d135f5e695e83b850a1e0cfe9685e90ace419a656668e07405a7e28d190014ade143328f8ad956f76a10bda1e078ec171f591bb102174f8945b62e0ba8beef85a8207bb5b1f9b6e486f5c1ecbbdf5686b34041b97e7a1a9afcab54cec4b9c9b6"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)

    def test_response_request_section_status(self):
        # pylint: disable=line-too-long
        message = UNiiResponseMessage(
            bytes.fromhex("d3f5659325030000c8320502001af6c258882c6a41117f248f8f"),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        _LOGGER.debug(message)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Alphatronics UNii Unit Tests
"""

import asyncio
import logging

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(filename)s:%(lineno)d %(message)s",
    level=logging.DEBUG,
)


def async_test(coro):
    """
    Creates an event loop wrapper around an async test case
    """

    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        try:
            return asyncio.run(coro(*args, **kwargs))
        finally:
            loop.close()

    return wrapper



# pylint: disable=R0801
# pylint: disable=missing-function-docstring
"""
Test UNii Command Data helper functions.
"""

import logging
import unittest

from unii.unii_command_data import bcd_encode, bit_position_to_numeric

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Test UNii Command Data helper function correctness.
    """

    def test_bit_position_to_numeric_1(self):
        result = bit_position_to_numeric(0b0000000000000001.to_bytes(2))
        self.assertListEqual(result, [1])

    def test_bit_position_to_numeric_2(self):
        result = bit_position_to_numeric(0b0000000000000010.to_bytes(2))
        self.assertListEqual(result, [2])

    def test_bit_position_to_numeric_3(self):
        result = bit_position_to_numeric(0b0000000000000011.to_bytes(2))
        self.assertListEqual(result, [1, 2])

    def test_bcd_encode(self):
        result = bcd_encode("00000000")
        self.assertEqual(result, 0x0000000000000000.to_bytes(8))

    def test_bcd_encode_12345678(self):
        result = bcd_encode("12345678")
        self.assertEqual(result, 0x1234567800000000.to_bytes(8))

    def test_bcd_encode_22170302(self):
        result = bcd_encode("22170302")
        self.assertEqual(result, 0x2217030200000000.to_bytes(8))



"""
To avoid a task disappearing mid-execution a reference to the task needs to be saved. The event
loop only keeps weak references to tasks. A task that isn't referenced elsewhere may get garbage
collected at any time, even before it's done. For reliable "fire-and-forget" background tasks,
gather them in a collection.

https://docs.python.org/3/library/asyncio-task.html#creating-tasks
"""

import asyncio
from typing import Set

background_tasks: Set[asyncio.Task] = set()


def save_task_reference(task: asyncio.Task) -> None:
    "To avoid a task disappearing mid-execution save a reference."
    # Add task to the set. This creates a strong reference.
    background_tasks.add(task)

    # To prevent keeping references to finished tasks forever, make each task remove its own
    # reference from the set after completion:
    task.add_done_callback(background_tasks.discard)



# pylint: disable=R0801
# pylint: disable=missing-function-docstring
"""
Test UNii Command Data helper functions.
"""

import logging
import unittest

from unii.unii_command_data import UNiiBypassMode, UNiiBypassUnbypassZoneInput

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Test UNii Command Data helper function correctness.
    """

    def test_bypass_input(self):
        data = UNiiBypassUnbypassZoneInput(UNiiBypassMode.USER_CODE, "123456", 1)
        self.assertEqual(data.mode, 0)
        self.assertEqual(data.code, "12345600")
        self.assertEqual(data.number, 1)
        self.assertEqual(data.to_bytes(), bytes.fromhex("0012345600000000000001"))



"""
Implementation of the Alphatronics UNii library
"""

try:
    from ._version import __version__
except ModuleNotFoundError:
    pass
from .sia_code import SIACode
from .unii import (
    DEFAULT_PORT,
    UNii,
    UNiiCommand,
    UNiiData,
    UNiiEncryptionError,
    UNiiFeature,
    UNiiLocal,
)
from .unii_command_data import (
    UNiiInputState,
    UNiiInputStatusRecord,
    UNiiSection,
    UNiiSectionArmedState,
    UNiiSectionStatusRecord,
    UNiiSensorType,
)



# pylint: disable=protected-access
# pylint: disable=R0801
"""
Tests unencrypted connection to Alphatronics UNii.
"""

import asyncio
import json
import logging
import unittest

from unii import UNiiLocal

from . import async_test

_SETTINGS_JSON = "settings.json"

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Unit test for unencrypted connecting to Alphatronics UNii.
    """

    _host = None
    _port: int = 6502

    def setUp(self):
        with open(_SETTINGS_JSON, encoding="utf8") as settings_file:
            settings = json.load(settings_file)
            self._host = settings.get("host")
            self._port = settings.get("unencrypted_port", self._port)

    def test_create_unii_object(self):
        """
        Test creating an Alphatronics UNii object
        """
        unii = UNiiLocal(self._host, self._port)
        self.assertIsNotNone(unii)

    @async_test
    async def test_connect(self):
        """
        Test connecting to Alphatronics UNii.
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port)
        try:
            result = await unii.connect()
            self.assertTrue(result, "Failed to connect to UNii")
        finally:
            await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_connect_unavailable_host(self):
        """
        Test connecting to Alphatronics UNii.
        """
        await asyncio.sleep(1)
        unii = UNiiLocal("127.0.0.1", self._port)
        try:
            result = await unii.connect()
            self.assertFalse(result, "Failed to connect to UNii")
        finally:
            await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_disconnect(self):
        """
        Test disconnectiong from Alphatronics UNii.
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port)
        await unii.connect()
        await asyncio.sleep(1)
        result = await unii.disconnect()
        self.assertTrue(result, "Failed to disconnect from UNii")

    @async_test
    async def test_poll_alive(self):
        """
        Test Poll Alive request.
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port)
        try:
            await unii.connect()
            await asyncio.sleep(3)
            result = await unii._poll_alive()
            self.assertTrue(result, "Failed to poll alive UNii")
        finally:
            await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_poll_alive_timeout(self):
        """
        Test Poll Alive timeout.

        To keep the connection alive (and NAT entries active) a poll message has to be sent every
        30 seconds if no other messages where sent during the last 30 seconds.

        The UNii takes care of sending this poll message and should keep the connection
        alive.
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port)
        try:
            await unii.connect()
            await asyncio.sleep(50)
            self.assertTrue(unii.connection.is_open, "Connection is not open")
        finally:
            await unii.disconnect()

    @async_test
    async def test_poll_alive_3x(self):
        """
        Test connecting to Alphatronics UNii
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port)
        try:
            await unii.connect()
            n = 3
            for i in range(n):
                await asyncio.sleep(3)
                _LOGGER.debug("Poll alive %i/%i", i + 1, n)
                result = await unii._poll_alive()
                with self.subTest(msg="Poll alive"):
                    self.assertTrue(result, "Failed to poll alive UNii")
                    _LOGGER.debug("Poll alive %i/%i success", i + 1, n)
        finally:
            await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_poll_alive_10x(self):
        """
        Test connecting to Alphatronics UNii
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port)
        try:
            await unii.connect()
            n = 10
            for i in range(n):
                await asyncio.sleep(1)
                _LOGGER.debug("Poll alive %i/%i", i + 1, n)
                result = await unii._poll_alive()
                with self.subTest(msg="Poll alive"):
                    self.assertTrue(result, "Failed to poll alive UNii")
                    _LOGGER.debug("Poll alive %i/%i success", i + 1, n)
        finally:
            await asyncio.sleep(1)
            await unii.disconnect()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Commands used by the Alphatronics UNii.
"""

from enum import IntEnum
from typing import Final


class UNiiCommand(IntEnum):
    """
    UNii Commands
    """

    # Virtual commands used within the library
    RELOAD_CONFIGURATION: Final = 0x010001
    REAUTHENTICATE: Final = 0x010002

    # Connection setup
    CONNECTION_REQUEST: Final = 0x0001
    CONNECTION_REQUEST_RESPONSE: Final = 0x0002
    CONNECTION_REQUEST_DENIED: Final = 0x0003
    NORMAL_DISCONNECT: Final = 0x0014

    # Operational commands
    POLL_ALIVE_REQUEST: Final = 0x0012
    POLL_ALIVE_RESPONSE: Final = 0x0013
    GENERAL_RESPONSE: Final = 0x0101
    EVENT_OCCURRED: Final = 0x0102
    RESPONSE_EVENT_OCCURRED: Final = 0x0103
    FLUSH_EVENT_BUFFER: Final = 0x0104
    INPUT_STATUS_CHANGED: Final = 0x0105
    REQUEST_INPUT_STATUS: Final = 0x0106
    DEVICE_STATUS_CHANGED: Final = 0x0107
    REQUEST_DEVICE_STATUS: Final = 0x0108
    CLEAR_ALARM_MEMORY: Final = 0x0109
    REQUEST_READY_TO_ARM_SECTIONS: Final = 0x0110
    RESPONSE_READY_TO_ARM_SECTIONS: Final = 0x0111
    REQUEST_ARM_SECTION: Final = 0x0112
    RESPONSE_ARM_SECTION: Final = 0x0113
    REQUEST_DISARM_SECTION: Final = 0x0114
    RESPONSE_DISARM_SECTION: Final = 0x0115
    REQUEST_SECTION_STATUS: Final = 0x0116
    RESPONSE_REQUEST_SECTION_STATUS: Final = 0x0117
    REQUEST_TO_BYPASS_AN_INPUT: Final = 0x0118
    RESPONSE_REQUEST_TO_BYPASS_AN_INPUT: Final = 0x0119
    REQUEST_TO_UNBYPASS_AN_INPUT: Final = 0x011A
    RESPONSE_REQUEST_TO_UNBYPASS_AN_INPUT: Final = 0x011B
    REQUEST_TO_SET_OUTPUT: Final = 0x011C
    RESPONSE_REQUEST_TO_SET_OUTPUT: Final = 0x011D
    REQUEST_OUTPUT_STATUS: Final = 0x011E
    RESPONSE_REQUEST_OUTPUT_STATUS: Final = 0x011F
    # REQUEST TO SET OUTPUT(S) WITH USER CODE: Final = 0X0120
    # RESPONSE REQUEST TO SET OUTPUT(S) WITH USER CODE: Final = 0X0121
    REQUEST_TO_RESET_ALARM: Final = 0x0122
    RESPONSE_REQUEST_TO_RESET_ALARM: Final = 0x0123
    EXIT_ENTRY_TIMER: Final = 0x0124
    INPUT_STATUS_UPDATE: Final = 0x0125

    # Configuration commands
    REQUEST_SECTION_ARRANGEMENT: Final = 0x0130
    RESPONSE_REQUEST_SECTION_ARRANGEMENT: Final = 0x0131
    REQUEST_GROUP_ARRANGEMENT: Final = 0x0132
    RESPONSE_REQUEST_GROUP_ARRANGEMENT: Final = 0x0133
    REQUEST_INPUT_ARRANGEMENT: Final = 0x0140
    RESPONSE_REQUEST_INPUT_ARRANGEMENT: Final = 0x0141
    REQUEST_EQUIPMENT_INFORMATION: Final = 0x0142
    RESPONSE_REQUEST_EQUIPMENT_INFORMATION: Final = 0x0143
    REQUEST_DATE_AND_TIME: Final = 0x0144
    RESPONSE_REQUEST_DATE_AND_TIME: Final = 0x0145
    WRITE_DATE_AND_TIME: Final = 0x0146
    RESPONSE_WRITE_DATE_AND_TIME: Final = 0x0147
    REQUEST_OUTPUT_ARRANGEMENT: Final = 0x0148
    RESPONSE_REQUEST_OUTPUT_ARRANGEMENT: Final = 0x0149

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name



"""
Classes for connecting to Alphatronics UNii security systems.

Only TCP/IP connections are supported, UDP connections are currently not implemented.
"""

import asyncio
import logging
import random
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Final

from .task_helper import save_task_reference
from .unii_command import UNiiCommand
from .unii_command_data import UNiiData
from .unii_message import UNiiMessageError, UNiiRequestMessage, UNiiResponseMessage

logger = logging.getLogger(__name__)

DEFAULT_PORT: Final = 6502


class UNiiConnectionError(Exception):
    """
    UNii Connection Error.

    When an error occurs while connecting to the Alphatronics UNii.
    """


# class UNiiSequenceNumberError(Exception):
#     """
#     UNii Sequence Number Error.
#
#     When the received RX Sequence Number does not match the sent TX Sequence Number.
#     """
#
#     def __init__(self, expected_sequence_number: int, received_sequence_number: int):
#         self._expected_sequence_number = expected_sequence_number
#         self._received_sequence_number = received_sequence_number
#
#     def __str__(self) -> str:
#         return (
#             f"Invalid sequence number. Expected {self._expected_sequence_number}, "
#             + "received {self._received_sequence_number}."
#         )


class UNiiConnection(ABC):
    """
    Abstract connection class on which the different connection types are build.
    """

    _MAX_LENGTH_OF_ANSWER: Final = 1500

    last_message_sent: datetime = datetime.now()
    last_message_received: datetime = datetime.now()
    _message_received_callback = None
    unique_id: str

    def set_message_received_callback(self, callback):
        """
        Sets the call back for handling Event Occurred messages.
        """
        self._message_received_callback = callback

    @abstractmethod
    async def connect(self):
        """
        Opens a connection to the Alphatronics Unii.
        """
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> bool:
        """
        Closes the connection to the Alphatronics UNii.
        """
        raise NotImplementedError

    @property
    def is_open(self) -> bool:
        """
        If the connection is open
        """
        raise NotImplementedError

    @property
    def is_encrypted(self) -> bool:
        """
        If the connection is open
        """
        raise NotImplementedError

    @abstractmethod
    async def send(
        self,
        command: UNiiCommand,
        data: UNiiData | None = None,
    ) -> int | None:
        """
        Constructs a message based on given command and data, writes this to the UNii and
        returns the TX Sequence ID or None if writing failed.
        """
        raise NotImplementedError


class UNiiTCPConnection(UNiiConnection):
    # pylint: disable=too-many-instance-attributes
    """
    Connection to an Alphatronics UNii over TCP/IP.
    """

    _reader: asyncio.StreamReader | None = None
    _writer: asyncio.StreamWriter | None = None
    _session_id: int | None = None
    _tx_sequence: int = -1
    _rx_sequence: int = 0

    _receive_task: asyncio.Task | None = None

    def __init__(
        self, host: str, port: int = DEFAULT_PORT, shared_key: bytes | None = None
    ):
        assert host is not None
        assert port is not None

        self._host = host
        self._port = port
        self._shared_key = shared_key
        self._writer_lock = asyncio.Lock()
        self.unique_id = host

    def __str__(self) -> str:
        return f"{self._host}:{self._port}"

    async def connect(self):
        try:
            if not self.is_open:
                logger.info("Connecting to %s", self)
                self._reader, self._writer = await asyncio.wait_for(
                    asyncio.open_connection(self._host, self._port), timeout=10
                )
                logger.info("Connected to %s", self)

                # Fixed session ID only one session
                self._session_id = 0xFFFF
                # Start with a random TX Sequence
                self._tx_sequence = random.randint(0, 65535)

                # Start the read coroutine
                self._receive_task = asyncio.create_task(self._receive_coroutine())
                save_task_reference(self._receive_task)
        except TimeoutError as ex:
            raise UNiiConnectionError("Open connection timeout") from ex
        except (
            OSError,
            ConnectionRefusedError,
            ConnectionResetError,
            asyncio.CancelledError,
        ) as ex:
            raise UNiiConnectionError(ex) from ex

    async def _close(self) -> bool:
        if self.is_open:
            try:
                async with self._writer_lock:
                    self._writer.close()
                    await self._writer.wait_closed()
            except ConnectionResetError as ex:
                logger.error(ex)

            self._writer = None
            self._reader = None

        if not self.is_open:
            logger.debug("Connection closed")
            return True

        logger.error("Failed to close connection")
        return False

    async def close(self) -> bool:
        if not await self._cancel_receive():
            return False

        return await self._close()

    @property
    def is_open(self) -> bool:
        """
        If the connection is open
        """
        return self._writer is not None

    @property
    def is_encrypted(self) -> bool:
        """
        If the connection is open
        """
        return self._shared_key is not None

    async def _cancel_receive(self) -> bool:
        if self._receive_task is not None and not (
            self._receive_task.done() or self._receive_task.cancelled()
        ):
            self._receive_task.cancel()
            try:
                await self._receive_task
            except asyncio.CancelledError:
                logger.debug("Receive task was cancelled")
                self._receive_task = None

        if self._receive_task is not None:
            logger.error("Failed to cancel receive task")
            logger.debug("Receive task: %s", self._receive_task)
            return False

        return True

    async def _receive_coroutine(self):
        while self._reader is not None:
            try:
                response = await asyncio.wait_for(self._reader.read(14), timeout=0.1)
                if len(response) < 13:
                    continue

                # Length
                packet_length = int.from_bytes(response[12:14])

                # Read remaining part of the message
                while len(response) < packet_length:
                    response += await self._reader.read(packet_length - len(response))
                    if len(response) < packet_length:
                        logger.debug("Response not yet fully received")

                message = UNiiResponseMessage(response, self._shared_key)
                if message.command not in [
                    # UNiiCommand.CONNECTION_REQUEST_RESPONSE,
                    UNiiCommand.POLL_ALIVE_RESPONSE,
                    UNiiCommand.INPUT_STATUS_CHANGED,
                    UNiiCommand.DEVICE_STATUS_CHANGED,
                    UNiiCommand.RESPONSE_REQUEST_INPUT_ARRANGEMENT,
                    UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION,
                ]:
                    logger.debug("Received: 0x%s", response.hex())
                    logger.debug("Received: %s", message)
                else:
                    logger.debug(
                        "Received: %i, %s", message.rx_sequence, message.command
                    )
                self.last_message_received = datetime.now()
                # logger.debug("Last message received: %s", self.last_message_sent)
                if message.rx_sequence != self._tx_sequence:
                    logger.warning(
                        "Invalid sequence number. Expected %i, received %i.",
                        self._tx_sequence,
                        message.rx_sequence,
                    )

                self._session_id = message.session_id
                self._rx_sequence = message.tx_sequence

                if self._message_received_callback is not None:
                    await self._message_received_callback(
                        message.rx_sequence, message.command, message.data
                    )
            except UNiiMessageError as ex:
                logger.error(ex)
            except ConnectionResetError as ex:
                logger.debug(ex)
                await self._close()
                break
            except asyncio.CancelledError:
                logger.debug("Receive coroutine was canceled")
                break
            except TimeoutError:
                pass
            except IndexError as ex:
                logger.error(ex)
            await asyncio.sleep(0.1)

        self._receive_task = None
        logger.debug("Receive coroutine stopped")

    async def send(
        self,
        command: UNiiCommand,
        data: UNiiData | None = None,
    ) -> int | None:
        if self.is_open:
            message = UNiiRequestMessage()
            message.session_id = self._session_id
            self._tx_sequence += 1
            message.tx_sequence = self._tx_sequence
            message.rx_sequence = self._rx_sequence
            message.command = command
            message.data = data
            # logger.debug("Sending: %s", message)
            # logger.debug("Sending: 0x%s", message.to_bytes(self._shared_key).hex())
            if self._writer is not None:
                try:
                    async with self._writer_lock:
                        self._writer.write(message.to_bytes(self._shared_key))
                        await self._writer.drain()
                        self.last_message_sent = datetime.now()
                        # logger.debug("Last message sent: %s", self.last_message_sent)

                    return message.tx_sequence
                except ConnectionResetError as ex:
                    await self.close()
                    raise UNiiConnectionError(str(ex)) from ex

        return None


# UDP connections are currently not implemented.
# class UNiiUDPConnection(UNiiConnection):
#     """
#     Connection to an Alphatronics UNii over UDP
#     """
#     pass



# pylint: disable=protected-access
# pylint: disable=R0801
"""
Test UNii message encryption.
"""

import logging
import unittest

from unii.unii_command import UNiiCommand
from unii.unii_message import UNiiRequestMessage, UNiiResponseMessage

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Test UNii message encryption correctness.
    """

    def test_message_encryption(self):
        # pylint: disable=line-too-long
        """
        Creates an encrypted connection request message and validates it's correctness
        """
        message = UNiiRequestMessage()
        message.session_id = 0xFFFF
        message.tx_sequence = 0x84AC0B7A
        message.rx_sequence = 0x00000000
        message.command = UNiiCommand.CONNECTION_REQUEST
        self.assertEqual(
            message.to_bytes(bytes.fromhex("31323334353637383930616263646566")).hex(),
            "ffff84ac0b7a000000000501002093458e6de62e1d5ea0e5281d5261f1845303",
        )

    def test_message_decryption(self):
        # pylint: disable=line-too-long
        """
        Creates an encrypted connection request message and validates it's correctness
        """
        message = UNiiResponseMessage(
            bytes.fromhex(
                "ffff84ac0b7a000000000501002093458e6de62e1d5ea0e5281d5261f1845303"
            ),
            bytes.fromhex("31323334353637383930616263646566"),
        )
        self.assertEqual(message.session_id, 0xFFFF)
        self.assertEqual(message.tx_sequence, 0x84AC0B7A)
        self.assertEqual(message.rx_sequence, 0x00000000)
        self.assertEqual(message.command, UNiiCommand.CONNECTION_REQUEST)
        self.assertIsNone(message.data)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
SIA Codes
"""

from enum import StrEnum
from typing import Final


class SIACode(StrEnum):
    """
    SIA Codes
    """

    AC_RESTORAL: Final = "AR"
    AC_TROUBLE: Final = "AT"
    BURGLARY_ALARM: Final = "BA"
    BURGLARY_BYPASS: Final = "BB"
    BURGLARY_CANCEL: Final = "BC"
    SWINGER_TROUBLE: Final = "BD"
    SWINGER_TROUBLE_RESTORE: Final = "BE"
    BURGLARY_TROUBLE_RESTORE: Final = "BJ"
    BURGLARY_RESTORAL: Final = "BR"
    BURGLARY_TROUBLE: Final = "BT"
    BURGLARY_UNBYPASS: Final = "BU"
    BURGLARY_VERIFIED: Final = "BV"
    BURGLARY_TEST: Final = "BX"
    CLOSING_DELINQUENT: Final = "CD"
    FORCED_CLOSING: Final = "CF"
    CLOSE_AREA: Final = "CG"
    FAIL_TO_CLOSE: Final = "CI"
    EARLY_TO_CLOSE: Final = "CK"
    CLOSING_REPORT: Final = "CL"
    AUTOMATIC_CLOSING: Final = "CP"
    REMOTE_CLOSING: Final = "CQ"
    CLOSING_KEYSWITCH: Final = "CS"
    LATE_TO_OPEN: Final = "CT"
    ACCESS_CLOSED: Final = "DC"
    ACCESS_DENIED: Final = "DD"
    DOOR_FORCED: Final = "DF"
    ACCESS_GRANTED: Final = "DG"
    ACCESS_DENIED_PASSBACK: Final = "DI"
    DOOR_LEFT_OPEN: Final = "DN"
    ACCESS_OPEN: Final = "DO"
    DOOR_RESTORAL: Final = "DR"
    REQUEST_TO_EXIT: Final = "DX"
    EXIT_ALARM: Final = "EA"
    EXPANSION_TAMPER_RESTORE: Final = "EJ"
    EXPANSION_MISSING: Final = "EM"
    EXPANSION_MISSING_RESTORE: Final = "EN"
    EXPANSION_RESTORAL: Final = "ER"
    EXPANSION_DEVICE_TAMPER: Final = "ES"
    EXPANSION_TROUBLE: Final = "ET"
    FIRE_ALARM: Final = "FA"
    FIRE_BYPASS: Final = "FB"
    FIRE_CANCEL: Final = "FC"
    FIRE_TROUBLE_RESTORE: Final = "FJ"
    FIRE_RESTORAL: Final = "FR"
    FIRE_TROUBLE: Final = "FT"
    FIRE_UNBYPASS: Final = "FU"
    HOLDUP_ALARM: Final = "HA"
    HOLDUP_BYPASS: Final = "HB"
    HOLDUP_TROUBLE_RESTORE: Final = "HJ"
    HOLDUP_RESTORAL: Final = "HR"
    HOLDUP_TROUBLE: Final = "HT"
    HOLDUP_UNBYPASS: Final = "HU"
    CONFIRMED_HOLDUP: Final = "HV"
    USER_CODE_TAMPER: Final = "JA"
    TIME_CHANGED: Final = "JT"
    LOCAL_PROGRAMMING: Final = "LB"
    MODEM_RESTORAL: Final = "LR"
    MODEM_TROUBLE: Final = "LT"
    LOCAL_PROGRAMMING_ENDED: Final = "LX"
    MEDICAL_ALARM: Final = "MA"
    MEDICAL_BYPASS: Final = "MB"
    MEDICAL_TROUBLE_RESTORE: Final = "MJ"
    MEDICAL_RESTORAL: Final = "MR"
    MEDICAL_TROUBLE: Final = "MT"
    MEDICAL_UNBYPASS: Final = "MU"
    PERIMETER_ARMED: Final = "NL"
    NETWORK_LINK_IP_RESTORE: Final = "NR"
    NETWORK_LINK_GPRS_RESTORE: Final = "NR"
    NETWORK_LINK_IP_FAIL: Final = "NT"
    NETWORK_LINK_GPRS_FAIL: Final = "NT"
    AUTOMATIC_OPENING: Final = "OA"
    OPEN_AREA: Final = "OG"
    EARLY_OPEN: Final = "OK"
    OPENING_REPORT: Final = "OP"
    OPENING_KEYSWITCH: Final = "OS"
    LATE_TO_CLOSE: Final = "OT"
    REMOTE_OPENING: Final = "OQ"
    DISARM_FROM_ALARM: Final = "OR"
    PANIC_ALARM: Final = "PA"
    PANIC_BYPASS: Final = "PB"
    PANIC_TROUBLE_RESTORE: Final = "PJ"
    PANIC_RESTORAL: Final = "PR"
    PANIC_TROUBLE: Final = "PT"
    PANIC_UNBYPASS: Final = "PU"
    RELAY_CLOSE: Final = "RC"
    REMOTE_RESET: Final = "RN"
    RELAY_OPEN: Final = "RO"
    AUTOMATIC_TEST: Final = "RP"
    POWERUP: Final = "RR"
    REMOTE_PROGRAM_SUCCESS: Final = "RS"
    DATA_LOST: Final = "RT"
    MANUAL_TEST: Final = "RX"
    TAMPER: Final = "TA"
    TAMPER_BYPASS: Final = "TB"
    TAMPER_RESTORAL: Final = "TR"
    TAMPER_UNBYPASS: Final = "TU"
    TEST_CALL: Final = "TX"
    UNTYPED_ALARM: Final = "UA"
    UNTYPED_BYPASS: Final = "UB"
    UNTYPED_TROUBLE_RESTORE: Final = "UJ"
    UNTYPED_RESTORAL: Final = "UR"
    UNTYPED_TROUBLE: Final = "UT"
    UNTYPED_UNBYPASS: Final = "UU"
    BELL_FAULT: Final = "YA"
    RF_JAM_RESTORAL: Final = "XH"
    RF_TAMPER_RESTORAL: Final = "XJ"
    READER_LOCKED: Final = "RL"
    READER_UNLOCKED: Final = "RG"
    KEYPAD_UNLOCKED: Final = "KG"
    RF_JAM_FAULT: Final = "XQ"
    RF_TAMPER: Final = "XS"
    COMMUNICATION_FAIL: Final = "YC"
    CHECKSUM_FAULT: Final = "YF"
    BELL_RESTORED: Final = "YH"
    COMMUNICATION_RESTORAL: Final = "YK"
    BATTERY_MISSING: Final = "YM"
    PSU_TROUBLE: Final = "YP"
    PSU_RESTORAL: Final = "YQ"
    BATTERY_RESTORAL: Final = "YR"
    COMMUNICATION_TROUBLE: Final = "YS"
    BATTERY_TROUBLE: Final = "YT"
    WATCHDOG_RESET: Final = "YW"
    SERVICE_REQUIRED: Final = "YX"
    SERVICE_COMPLETED: Final = "YZ"
    # Special SIA events
    USER_DURESS: Final = "HA"
    USER_DURESS_RESTORE: Final = "HR"
    ENET_PANIC_ALARM: Final = "PA"
    ENET_PANIC_RESTORAL: Final = "PR"
    USER_PANIC_ALARM: Final = "PA"
    ENET_FIRE_ALARM: Final = "FA"
    ENET_FIRE_RESTORAL: Final = "FR"
    ENET_MEDICAL_ALARM: Final = "MA"
    ENET_MEDICAL_RESTORAL: Final = "MR"
    MDT_PANIC: Final = "PA"
    MDT_TILT: Final = "MA"
    MDT_BELT_CLIP: Final = "HA"
    MDT_PANIC_RESTORE: Final = "PR"
    MDT_TILT_RESTORE: Final = "MR"
    MDT_BELT_CLIP_RESTORE: Final = "HR"
    RPA_PANIC: Final = "PA"
    RPA_PANIC_RESTORE: Final = "PR"
    RPA_HOLDUP: Final = "HA"
    RPA_HOLDUP_RESTORE: Final = "HR"
    USER_CODE_CHANGE: Final = "JV"

    # Non-standard SIA codes for zone state reporting
    ZONE_OPEN: Final = "ZO"
    ZONE_CLOSE: Final = "ZC"
    ZONE_SHORT: Final = "ZX"
    ZONE_DISCON: Final = "ZD"
    ZONE_MASKED: Final = "ZM"
    ZONE_WALKED: Final = "TP"
    WALKTEST_START: Final = "ZK"
    WALKTEST_END: Final = "TC"
    ZONE_LOW_BATT: Final = "XT"
    ZONE_LOW_BATTERY_RESTORAL: Final = "XR"

    # Other non-standard SIA codes
    CAMERA_ONLINE: Final = "CU"
    CAMERA_OFFLINE: Final = "CV"
    ALERT_CLOSE: Final = "SD"
    ALERT_REOPEN: Final = "SO"
    XBUS_ALERT_CLOSE: Final = "NB"
    XBUS_ALERT_REOPEN: Final = "NO"
    UNKNOWN_CARD: Final = "AU"
    USER_ACCESSING: Final = "JP"
    USER_ACCESSING_END: Final = "ZG"
    LOW_VOLTAGE: Final = "XD"
    LOW_VOLTAGE_RESTORAL: Final = "XG"
    DEEP_CHARGE: Final = "XK"
    LOCKED_OUT: Final = "WW"

    def __repr__(self) -> str:
        return self.name
